/* -*- mode: c ; c-file-style: "canonware-c-style" -*-
 ******************************************************************************
 *
 * Copyright (C) 1996-2005 Jason Evans <jasone@canonware.com>.
 * All rights reserved.
 * 
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 * 1. Redistributions of source code must retain the above copyright
 *    notice(s), this list of conditions and the following disclaimer
 *    unmodified other than the allowable addition of one or more
 *    copyright notices.
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice(s), this list of conditions and the following disclaimer in
 *    the documentation and/or other materials provided with the
 *    distribution.
 * 
 * THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDER(S) ``AS IS'' AND ANY
 * EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
 * IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR
 * PURPOSE ARE DISCLAIMED.  IN NO EVENT SHALL THE COPYRIGHT HOLDER(S) BE
 * LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR
 * CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF
 * SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR
 * BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY,
 * WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE
 * OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE,
 * EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 *
 ******************************************************************************
 *
 * Version: Onyx 5.1.2
 *
 ******************************************************************************/

typedef struct cw_nxoe_string_s cw_nxoe_string_t;

struct cw_nxoe_string_s
{
    cw_nxoe_t nxoe;

#ifdef CW_THREADS
    /* Access is locked if this object has the locking bit set.  Indirect
     * strings aren't locked, but their parents are. */
    cw_mtx_t lock;
#endif

    union
    {
	struct
	{
	    cw_nxoe_string_t *string;
	    uint32_t beg_offset;
	    uint32_t len;
	} i;
	struct
	{
	    char *str;
	    uint32_t len;
	    uint32_t alloc_len;
	} s;
    } e;
};

#ifndef CW_USE_INLINES
bool
nxoe_l_string_delete(cw_nxoe_t *a_nxoe, uint32_t a_iter);

cw_nxoe_t *
nxoe_l_string_ref_iter(cw_nxoe_t *a_nxoe, bool a_reset);
#endif

#if (defined(CW_USE_INLINES) || defined(CW_NXO_STRING_C_))
CW_INLINE bool
nxoe_l_string_delete(cw_nxoe_t *a_nxoe, uint32_t a_iter)
{
    cw_nxoe_string_t *string;

    string = (cw_nxoe_string_t *) a_nxoe;

    cw_check_ptr(string);
    cw_dassert(string->nxoe.magic == CW_NXOE_MAGIC);
    cw_assert(string->nxoe.type == NXOT_STRING);

    if (string->nxoe.indirect == false && string->e.s.alloc_len > 0)
    {
	nxa_free(string->e.s.str, string->e.s.alloc_len);
    }

#ifdef CW_THREADS
    if (string->nxoe.locking && string->nxoe.indirect == false)
    {
	mtx_delete(&string->lock);
    }
#endif

    nxa_free(string, sizeof(cw_nxoe_string_t));

    return false;
}

CW_INLINE cw_nxoe_t *
nxoe_l_string_ref_iter(cw_nxoe_t *a_nxoe, bool a_reset)
{
    cw_nxoe_t *retval;
    cw_nxoe_string_t *string;
    /* Used for remembering the current state of reference iteration.  This
     * function is only called by the garbage collector, so using a static
     * variable works fine. */
    static uint32_t ref_iter;

    string = (cw_nxoe_string_t *) a_nxoe;

    if (a_reset)
    {
	ref_iter = 0;
    }

    if (a_nxoe->indirect == false)
    {
	retval = NULL;
    }
    else if (ref_iter == 0)
    {
	retval = (cw_nxoe_t *) string->e.i.string;
	ref_iter++;
    }
    else
    {
	retval = NULL;
    }

    return retval;
}
#endif /* (defined(CW_USE_INLINES) || defined(CW_NXO_STRING_C_)) */
